/*
 *  compressed.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "packet_private.h"
#include "simplepgp.h"
#include "compressed.h"
#include "zlib.h"

uint8_t spgp_zlib_decompress_buffer(uint8_t *inbuf, uint32_t inlen,
                                    uint8_t **outbuf, uint32_t *outlen,
                                    uint8_t algo) {
  uint32_t maxsize;
  z_stream s;
  uint8_t *tmpbuf;
  int wbits;
  int err;
  int i;
  
  if (NULL == inbuf || inlen == 0 || NULL == outbuf || NULL == outlen)
    RAISE(INVALID_ARGS);
  
  maxsize = inlen * 100;
  
  *outbuf = malloc(maxsize);
  if (NULL == *outbuf) RAISE(OUT_OF_MEMORY);
  
  s.zalloc = Z_NULL;
  s.zfree = Z_NULL;
  s.next_in = inbuf;
  s.avail_in = inlen;
  s.next_out = *outbuf;
  s.avail_out = maxsize;

  if (algo == COMPRESSION_ZIP) wbits = -15;
  else wbits = 15;
  if (inflateInit2(&s, wbits) != Z_OK) RAISE(ZLIB_ERROR);

  for (i = s.total_out; i < maxsize; i++)
    (*outbuf)[i] = 0x55;
  
  LOG_PRINT("Inflating up to %u bytes\n", maxsize);
  while ((err = inflate(&s, Z_NO_FLUSH)) != Z_STREAM_END) {
    if (err != Z_OK) RAISE(ZLIB_ERROR);
    if (s.avail_in == 0) break; // Done
    // If we're here, our output buffer isn't large enough
    maxsize <<= 1; // double size
    tmpbuf = *outbuf;
    *outbuf = realloc(*outbuf, maxsize);
    if (NULL == *outbuf) {
      free(tmpbuf);
      RAISE(OUT_OF_MEMORY);
    }
    s.next_out = *outbuf + s.total_out;
    for (i = s.total_out; i < maxsize; i++)
      (*outbuf)[i] = 0x55;
    s.avail_out = maxsize - s.total_out;
    LOG_PRINT("Grew to up to %u bytes\n", maxsize);
  }
  LOG_PRINT("Total inflated bytes: %lu\n", s.total_out);
  *outlen = s.total_out;
  
  if (inflateEnd(&s) != Z_OK) RAISE(ZLIB_ERROR);
  
  return 0;
}

uint8_t spgp_parse_compressed_packet(uint8_t *msg, 
                                     uint32_t *idx, 
                                     uint32_t length, 
                                     spgp_packet_t *pkt) {
  int algo;
  spgp_packet_t *pkts;
  uint8_t *decomp;
  uint32_t decomp_len;
  uint32_t didx;


  if (NULL == msg || NULL == idx || length == 0 || NULL == pkt)
    RAISE(INVALID_ARGS);
     
  algo = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  switch (algo) {
    case 1:
      LOG_PRINT("ZIP compressed packet\n");
      spgp_zlib_decompress_buffer(msg+*idx, pkt->header->contentLength,
                                  &decomp, &decomp_len, algo);
      break;
    case 2:
      LOG_PRINT("ZLIB compressed packet\n");
      spgp_zlib_decompress_buffer(msg+*idx, pkt->header->contentLength,
                                  &decomp, &decomp_len, algo);
      break;
    default:
      LOG_PRINT("Unsupported packet compression: %u\n", algo);
      RAISE(FORMAT_UNSUPPORTED);
  }
  
  if (NULL == decomp) RAISE(DECRYPT_FAILED);
  
  // Decode all the packets in this compressed packet        
  didx = 0;
  pkts = spgp_packet_decode_loop(decomp, &didx, decomp_len);
  if (NULL == pkts) RAISE(INCOMPLETE_PACKET);
  
  // Add packets to the current chain
  pkt->next = pkts;
  pkts->prev = pkt;
  
  free(decomp);
  decomp = NULL;
  
  // Progress index through current chain
  *idx += pkt->header->contentLength;
 
  return 0;
}
