/*
 *  header.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "simplepgp.h"
#include "packet_private.h"
#include "header.h"


uint8_t spgp_parse_header(uint8_t *msg, uint32_t *idx, 
                          uint32_t length, spgp_packet_t *pkt) {
  uint8_t i;
  
  if (NULL == msg || NULL == pkt || NULL == idx || length == 0)
    RAISE(INVALID_ARGS);

  // Allocate a header
  if (pkt->header == NULL) {
    LOG_PRINT("Allocating header.\n");
    pkt->header = malloc(sizeof(*(pkt->header)));
    if (pkt->header == NULL)
      RAISE(OUT_OF_MEMORY);
    memset(pkt->header, 0, sizeof(*(pkt->header)));
  }
  
  // Header points to its parent packet
  pkt->header->parent = pkt;
  
  // The first byte is the 'tag byte', which tells us the packet type.
  pkt->header->rawTagByte = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  LOG_PRINT("TAG BYTE: 0x%.2X\n", pkt->header->rawTagByte);
  
  // Validate tag byte -- top bit always set
  if (!(pkt->header->rawTagByte & 0x80))
    RAISE(INVALID_HEADER);
    
  // Second-MSB tells us if this is new or old format header
  pkt->header->isNewFormat = pkt->header->rawTagByte & 0x40;
  
  // Read the packet type out of the tag byte
  if (pkt->header->isNewFormat)
    pkt->header->type = pkt->header->rawTagByte & 0x1F;
  else // old style
    pkt->header->type = (pkt->header->rawTagByte >> 2) & 0x0F;
  LOG_PRINT("TYPE: 0x%.2X\n", pkt->header->type);
  
  // Read the length of the packet's contents.  In old packets, the length
  // is encoded into the tag byte. 
  if (pkt->header->isNewFormat == 0) {
    switch (pkt->header->rawTagByte & 0x03) {
      case 0: pkt->header->headerLength = 2; break;
      case 1: pkt->header->headerLength = 3; break;
      case 2: pkt->header->headerLength = 5; break;
      default: 
        // "indeterminate length" packet
        LOG_PRINT("Indeterminate length packet\n");
        pkt->header->headerLength = 1;
        pkt->header->contentLength = length-*idx-1;
    }
    for (i = 0; i < pkt->header->headerLength - 1; i++) {
      pkt->header->contentLength <<= 8;
      pkt->header->contentLength += msg[*idx];
      SAFE_IDX_INCREMENT(*idx, length);
    }
  }
  // In new packets, the length is encoded over a variable number of bytes, 
  // with the range of the first byte determining total number of bytes.
  else { // This is new style packet.
    pkt->header->contentLength = 
      spgp_new_header_length(msg+*idx, 
                             &(pkt->header->headerLength),
                             &(pkt->header->isPartial));
    *idx += pkt->header->headerLength - 2;
    SAFE_IDX_INCREMENT(*idx, length);
  }
  
  LOG_PRINT("LENGTH: %u\n", pkt->header->contentLength);
  
  return 0;
}

uint32_t spgp_new_header_length(uint8_t *header, 
                                uint8_t *header_len,
                                uint8_t *is_partial) {
  uint32_t content;
  uint8_t len[4];
  uint8_t i = 0;
  
  if (NULL == header || NULL == header_len) RAISE(INVALID_ARGS);
  
  *is_partial = 0; // default to known length
  
  len[0] = header[i];
  if (len[0] <= 191) { // 1-byte length
    *header_len = 2;
    content = len[0];
  }
  else if (len[0] > 191 && len[0] <= 223) { // 2-byte length
    *header_len = 3;
    len[1] = header[i+1]; 
    content = ((len[0]-192)<<8) | (len[1] + 192);
  }
  else if (len[0] == 255) { // 5-byte length
    *header_len = 5;
    len[0] = header[i+1];
    len[1] = header[i+2];
    len[2] = header[i+3];
    len[3] = header[i+4];
    content = (len[0]<<24) | (len[1]<<16) | (len[2]<<8) | len[3];
  }
  else {
    // indeterminate length
    LOG_PRINT("Partial length header!\n");
    *header_len = 2;
    *is_partial = 1;
    content = 1 << (len[0] & 0x1F);
  }
  return content;
}
