/*
 *  packet.c
 *  libsimplepgp
 *
 *  Created by Trevor Bentley on 11/1/11.
 *
 *  Copyright 2011 Trevor Bentley
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

#include "simplepgp.h"
#include "packet_private.h"
#include "keychain.h"
#include "util.h"
#include "mpi.h"
#include "pubkey.h"
#include "seckey.h"
#include "header.h"
#include "session.h"
#include "signature.h"
#include "compressed.h"

#include "gcrypt.h"

#include <wchar.h>
#include <locale.h>
#include <string.h>



/**********************************************************************
**
** Extern variables
**
***********************************************************************/

pthread_mutex_t spgp_mtx;


/**********************************************************************
**
** Static function prototypes
**
***********************************************************************/

static uint8_t spgp_parse_user_id(uint8_t *msg, uint32_t *idx, 
                                  uint32_t length, spgp_packet_t *pkt);
                                                                                                   
static uint8_t spgp_parse_encrypted_packet(uint8_t *msg, 
                                           uint32_t *idx, 
                                            uint32_t *length, 
                                           spgp_packet_t *pkt);
             
static uint8_t spgp_parse_literal_packet(uint8_t *msg, 
                                         uint32_t *idx, 
                                          uint32_t length, 
                                         spgp_packet_t *pkt);
                                                                                   
                                                           


/**********************************************************************
**
** External function definitions
**
***********************************************************************/
#pragma mark External Function Definitions

uint8_t spgp_init(void) {
  if (pthread_mutex_init(&spgp_mtx, NULL)) return -1;
  if (spgp_keychain_init()) return -1;
  return 0;
}

uint8_t spgp_close(void) {
  spgp_packet_t *chain = NULL;
  if (spgp_keychain_is_valid()) {
    spgp_keychain_iter_start();
    while ((chain = spgp_keychain_iter_next()) != NULL) {
      spgp_free_packet(&chain);
    }
    spgp_keychain_iter_end();
    spgp_keychain_free();
  }

  pthread_mutex_destroy(&spgp_mtx);
  return 0;
}

spgp_packet_t* spgp_packet_decode_loop(uint8_t *message, 
                                       uint32_t *idx, 
                                       uint32_t length) {
  spgp_packet_t *head = NULL;
  spgp_packet_t *pkt = NULL;

  // There must be at least one packet, yeah?
  head = malloc(sizeof(*head));
  if (NULL == head) RAISE(OUT_OF_MEMORY);
  memset(head, 0, sizeof(*head));
  pkt = head;
  
  // Loop to decode every packet in message
  while (*idx < length-1) {
     // Every packet starts with a header
    spgp_parse_header(message, idx, length, pkt);
    if (!pkt->header) RAISE(FORMAT_UNSUPPORTED);
    
    // Decode packet contents based on the type marked in its header
    switch (pkt->header->type) {
      case PKT_TYPE_USER_ID:
        spgp_parse_user_id(message, idx, length, pkt);
        break;
      case PKT_TYPE_PUBLIC_KEY:
      case PKT_TYPE_PUBLIC_SUBKEY:
        spgp_parse_public_key(message, idx, length, pkt);
        if (spgp_keychain_add_packet(pkt) != 0) RAISE(KEYCHAIN_ERROR);
        break;
      case PKT_TYPE_SECRET_KEY:
      case PKT_TYPE_SECRET_SUBKEY:
        spgp_parse_secret_key(message, idx, length, pkt);
        break;
      case PKT_TYPE_SESSION:
        spgp_parse_session_packet(message, idx, length, pkt);
        break;
      case PKT_TYPE_SYM_ENC_INT_DATA:
        spgp_parse_encrypted_packet(message, idx, &length, pkt);
        break;
      case PKT_TYPE_COMPRESSED_DATA:
        spgp_parse_compressed_packet(message, idx, length, pkt);
        break;
      case PKT_TYPE_LITERAL_DATA:
        spgp_parse_literal_packet(message, idx, length, pkt);
        break;
      case PKT_TYPE_SIGNATURE:
        spgp_parse_signature_packet(message, idx, length, pkt);
        break;
      default:
        LOG_PRINT("WARNING: Unsupported packet type %u\n", pkt->header->type);
        // Increment to next packet.  We add the contentLength, but subtract
        // one parse_header() left us on the first byte of content.
        if (*idx + pkt->header->contentLength - 1 < length)
          *idx = *idx + pkt->header->contentLength - 1;
        break;
    }
    
    // If we're at the end of the buffer, we're done
    if (*idx >= length-1) break;
        
    // A packet can contain other packets -- if such a thing was just decoded,
    // new packets have already been added to the list.  Progress until we
    // get to the end of the packet list.
    while (pkt->next != NULL) pkt = pkt->next;
        
    // Allocate space for another packet
    pkt->next = malloc(sizeof(*pkt->next));
    if (NULL == pkt->next) RAISE(OUT_OF_MEMORY);
    memset(pkt->next, 0, sizeof(*(pkt->next)));
    pkt->next->prev = pkt; // make backwards pointer
    pkt = pkt->next;
    
    // Packet parser increments to it's own last byte.  Need one more to get
    // to the next packet's first byte. 
    SAFE_IDX_INCREMENT(*idx, length);
  }

  return head;
}

spgp_packet_t *spgp_decode_message(uint8_t *message, uint32_t length) {
  spgp_packet_t *head = NULL;
  uint32_t idx = 0;
    
  switch (setjmp(exception)) {
    case 0:
      break; /* Run logic */
    /* Below here are exceptions */
    default:
      LOG_PRINT("Exception (0x%x)\n",_spgp_err);
      spgp_free_packet(&head);
      goto end;
  }

  if (NULL == message || 0 == length) {
    RAISE(INVALID_ARGS);
  }
  
  head = spgp_packet_decode_loop(message, &idx, length);

  end:
  return head;
}

char *spgp_get_literal_data(spgp_packet_t *msg, uint32_t *datalen,
                            char **filename, uint32_t *filenamelen) {
  spgp_packet_t *cur = msg;
  
  if (setjmp(exception)) {
      LOG_PRINT("Exception (0x%x)\n",_spgp_err);
      goto end;
  }
  
  if (NULL == msg || NULL == datalen || 
      NULL == filename || NULL == filenamelen) 
    RAISE(INVALID_ARGS);
  
  while (cur) {
    if (cur->header->type == PKT_TYPE_LITERAL_DATA) {
      *datalen = cur->c.literal->dataLen;
      *filenamelen = cur->c.literal->filenameLen;
      *filename = cur->c.literal->filename;
      return cur->c.literal->data;
    }
    cur = cur->next;
  }
  
  end:
  return NULL;
}


void spgp_free_packet(spgp_packet_t **pkt) {
  spgp_mpi_t *curMpi, *nextMpi;
  
  if (pkt == NULL)
    return;
  if (*pkt == NULL)
    return;
  
  //LOG_PRINT("Freeing packet: %p\n", pkt);
  
  // Recursively call on the next packet before freeing parent.
  if ((*pkt)->next) spgp_free_packet(&((*pkt)->next));
  (*pkt)->next = NULL;
  
  // Release memory allocated for secret key fields
  if (((*pkt)->header->type == PKT_TYPE_SECRET_KEY ||
      (*pkt)->header->type == PKT_TYPE_SECRET_SUBKEY) &&
      (*pkt)->c.secret != NULL) {
    if ((*pkt)->c.secret->pub.mpiCount > 0) {
      curMpi = (*pkt)->c.secret->pub.mpiHead;
      while (curMpi) {
        nextMpi = curMpi->next;
        if (curMpi->data) free(curMpi->data);
        free(curMpi);
        curMpi = nextMpi;
      }
      (*pkt)->c.secret->pub.mpiHead = NULL;
      (*pkt)->c.secret->pub.mpiCount = 0;
    }
    if ((*pkt)->c.secret->pub.fingerprint) {
      free((*pkt)->c.secret->pub.fingerprint);
    }
    if ((*pkt)->c.secret->encryptedData) {
      free((*pkt)->c.secret->encryptedData);
    }
    if ((*pkt)->c.secret->s2kSalt) {
      free((*pkt)->c.secret->s2kSalt);
      (*pkt)->c.secret->s2kSalt = NULL;
    }
    if ((*pkt)->c.secret->key) {
      free((*pkt)->c.secret->key);
      (*pkt)->c.secret->key = NULL;
    }
    if ((*pkt)->c.secret->iv) {
      free((*pkt)->c.secret->iv);
      (*pkt)->c.secret->iv = NULL;
    }
    free((*pkt)->c.secret);
    (*pkt)->c.secret = NULL;
  }
  
  else if (((*pkt)->header->type == PKT_TYPE_PUBLIC_KEY ||
            (*pkt)->header->type == PKT_TYPE_PUBLIC_SUBKEY) &&
            (*pkt)->c.pub != NULL) {
    if ((*pkt)->c.pub->mpiCount > 0) {
      curMpi = (*pkt)->c.pub->mpiHead;
      while (curMpi) {
        nextMpi = curMpi->next;
        if (curMpi->data) free(curMpi->data);
        free(curMpi);
        curMpi = nextMpi;
      }
      (*pkt)->c.pub->mpiHead = NULL;
      (*pkt)->c.pub->mpiCount = 0;
    }      
    if ((*pkt)->c.pub->fingerprint) {
      free((*pkt)->c.pub->fingerprint);
    }
    free((*pkt)->c.pub);
    (*pkt)->c.pub = NULL;
  }
  
  else if ((*pkt)->header->type == PKT_TYPE_USER_ID &&
           (*pkt)->c.userid->data != NULL) {
    free((*pkt)->c.userid->data);
    (*pkt)->c.userid->data = NULL;
    
    free((*pkt)->c.userid);
    (*pkt)->c.userid = NULL;
  }
  
  else if ((*pkt)->header->type == PKT_TYPE_SESSION &&
           (*pkt)->c.session != NULL) {
    if ((*pkt)->c.session->key) {
      free((*pkt)->c.session->key);
      (*pkt)->c.session->key = NULL;
    }
    if ((*pkt)->c.session->mpi1) {
      free((*pkt)->c.session->mpi1->data);
      free((*pkt)->c.session->mpi1);
      (*pkt)->c.session->mpi1 = NULL;
    }
    if ((*pkt)->c.session->mpi2) {
      free((*pkt)->c.session->mpi2->data);
      free((*pkt)->c.session->mpi2);
      (*pkt)->c.session->mpi2 = NULL;
    }
    free((*pkt)->c.session);
    (*pkt)->c.session = NULL;
  }
  
  else if ((*pkt)->header->type == PKT_TYPE_LITERAL_DATA &&
           (*pkt)->c.literal != NULL) {
    if ((*pkt)->c.literal->filename) {
      free((*pkt)->c.literal->filename);
      (*pkt)->c.literal->filename = NULL;
    }
    if ((*pkt)->c.literal->data) {
      free((*pkt)->c.literal->data);
      (*pkt)->c.literal->data = NULL;
    }
    free((*pkt)->c.literal);
    (*pkt)->c.literal = NULL;
  }
  else if ((*pkt)->header->type == PKT_TYPE_SIGNATURE &&
           (*pkt)->c.signature != NULL) {
    curMpi = (*pkt)->c.signature->mpiHead;
    while (curMpi) {
      nextMpi = curMpi->next;
      if (curMpi->data) free(curMpi->data);
      free(curMpi);
      curMpi = nextMpi;
    }
    free((*pkt)->c.signature);
    (*pkt)->c.signature = NULL;
  }
    
  // release header
  if ((*pkt)->header) {
    free((*pkt)->header);
    (*pkt)->header = NULL;
  }
  
  // release packet
  free(*pkt);

  *pkt = NULL;
}


/**********************************************************************
**
** Static function definitions
**
***********************************************************************/
#pragma mark Static Function Definitions


/**
 * \brief Parse User ID packet out of message
 *
 * |idx| will be updated to point ot last byte of the User ID,
 * |pkt| will have its c.userid fields filled in.  This function
 * dynamically allocates memory for fields.
 *
 * The User ID is just a short strong with name, e-mail address, or both
 * of the key holder.  There can be (and often are) multiple in a message.
 * String is in UTF-8 format.
 *
 * It is the caller's responsibility to free the structure and its data
 * when finished.  This is handled in spgp_free_packet().
 *
 * \param msg Message to parse
 * \param idx Index to current location of |msg|, updated to point to end
 *            of header when function returns.
 * \param length Length of |msg|.
 * \param pkt PGP packet structure to fill with data.
 * \return 0 on success, non-zero on failure.
 */
static uint8_t spgp_parse_user_id(uint8_t *msg, uint32_t *idx, 
                                  uint32_t length, spgp_packet_t *pkt) {
  spgp_userid_pkt_t *userid;

  LOG_PRINT("Parsing user id.\n");

  // Make sure we have enough bytes remaining for the copy
  if (length - *idx < pkt->header->contentLength) RAISE(BUFFER_OVERFLOW);
  
  // Allocate userid field in packet
  pkt->c.userid = malloc(sizeof(*(pkt->c.userid)));
  if (NULL == pkt->c.userid) RAISE(OUT_OF_MEMORY);
  userid = pkt->c.userid;
  
  // Allocate space for buffer, plus one byte for NUL terminator
  userid->data = malloc(sizeof(*(userid->data))*pkt->header->contentLength + 1);
  if (NULL == userid->data) RAISE(OUT_OF_MEMORY);
  
  // Copy bytes from input to structure, and add a NUL terminator
  memcpy(userid->data, msg+*idx, pkt->header->contentLength);
  userid->data[pkt->header->contentLength] = '\0';
  *idx += pkt->header->contentLength - 1;

  setlocale(LC_CTYPE, "en_US.UTF-8");
  fwprintf(stderr, L"USER ID: %s\n", pkt->c.userid->data);
  
  return 0;                                     
}


/**
 * \brief Parse encrypted data packet out of message
 *
 * The encrypted data in |msg| is decrypted IN PLACE.  |idx| is incremented
 * during decryption, but is reset back to the first packet of decrypted
 * data when this function exits.  This allows the processing loop to continue
 * running over the decrypted packets.
 *
 * "Partial packets" are handled by copying decrypted data down over the
 * partial packet headers.  This may be slow for very large packets.
 *
 * \param msg Message to parse
 * \param idx Index to current location of |msg|, reset to first decrypted
 *            packet byte when finished.
 * \param length Length of |msg|.
 * \param pkt PGP packet structure to fill with data.
 * \return 0 on success, non-zero on failure.
 */
static uint8_t spgp_parse_encrypted_packet(uint8_t *msg, 
                                           uint32_t *idx, 
                                            uint32_t *length, 
                                           spgp_packet_t *pkt) {
  spgp_packet_t *session_pkt;
  spgp_session_pkt_t *session;
  gcry_cipher_hd_t cipher_hd;
  gcry_error_t err;
  int version;
  unsigned long blksize;
  uint32_t encbytes;
  uint32_t startidx;
  uint8_t headerlen;
  uint8_t is_done;
  uint8_t is_partial;
  
  if (NULL == msg || NULL == idx || *length == 0 || NULL == pkt)
    RAISE(INVALID_ARGS);
    
  version = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, *length);
  
  // As of this writing, only version 1 exists
  if (version != 1) RAISE(FORMAT_UNSUPPORTED);
  
  session_pkt = spgp_find_session_packet(pkt);
  if (NULL == session_pkt) {
    LOG_PRINT("No session key found!\n");
    RAISE(DECRYPT_FAILED);
  }
  session = session_pkt->c.session;
  
  startidx = *idx;
  is_done = 0;
  is_partial = pkt->header->isPartial;
  
  // Drop 1 from contentLength to account for version
  encbytes = pkt->header->contentLength - 1;

  err = gcry_cipher_open (&cipher_hd, 
          spgp_pgp_to_gcrypt_symmetric_algo(session->symAlgo),
          GCRY_CIPHER_MODE_CFB,
          (GCRY_CIPHER_SECURE | GCRY_CIPHER_ENABLE_SYNC |
          GCRY_CIPHER_ENABLE_SYNC));
  blksize = spgp_iv_length_for_symmetric_algo(session->symAlgo);
  err |= gcry_cipher_setkey(cipher_hd, session->key, session->keylen);
  err |= gcry_cipher_setiv(cipher_hd, 0, blksize);
  if (err) RAISE(GCRY_ERROR);
  
  // Since data packets can have partial length, we loop and decrypt
  // here until the whole blasted thing is decrypted.
  while (!is_done) {
    err = gcry_cipher_decrypt(cipher_hd, 
                              msg+*idx, 
                              encbytes, 
                              NULL, 
                              0);
    if (err) RAISE(GCRY_ERROR);
    
    *idx += encbytes - 1; // increment to last byte of data
    
    // Check if we're done
    if (!(is_partial)) {
      is_done = 1;
      continue;
    }
    
    // We are processing a partial packet, so figure out next length
    SAFE_IDX_INCREMENT(*idx, *length); // inc to first byte of header
    encbytes = spgp_new_header_length(msg+*idx, 
                                       &(headerlen),
                                      &(is_partial));
    LOG_PRINT("%u more bytes\n", encbytes);

    // Fuck everything and everyone.  Partial lengths are the enemy of
    // gentlemen.  We shall kill it brutishly, by moving everything down.
    // God help you if your buffer is massive.
    //
    // The issue here is that a packet can technically be massive -- like
    // exabytes -- and that's still valid by the OpenPGP spec.  You can't
    // depend on the ability to create temporary buffers to stick this stuff
    // in.  But if you leave it in the original buffer, the sub-packets won't
    // know if there are "partial body length" headers jammed in the middle of
    // their data.  
    //
    // The only proper solution is some sort of dynamic stream data structure,
    // but it's too late for that now.  memmove() will have to do.  If your
    // packet is gigabytes long, this is going to take a while.
    memmove(msg+*idx, msg+*idx+headerlen-1, *length-(*idx+headerlen-1));
    *length -= headerlen-1;
  }

  // Validate decryption with PGP's MDC doo-hickey.  
  if (memcmp(msg+startidx+blksize-2, msg+startidx+blksize, 2) != 0) {
    LOG_PRINT("Decrypted data block fails validation!\n");
    RAISE(DECRYPT_FAILED);
  }
  LOG_PRINT("Decrypt succeeded.\n");

  // At this point, msg has been decrypted in place and now contains
  // a bunch of packets.  Since it was decoded in place, and since we're
  // already in the middle of a packet decode loop, we can just teleport
  // idx back to the beginning of the data and exit.
  //
  // Packet data starts blocksize+2 bytes above decrypted data.  One block
  // of random data, and 2 extra bytes of verification.
  //
  // Subtract 1 because startidx is the first byte of the decrypted packet, 
  // but the packet parser loop expects us to end on the last byte of the 
  // previous packet.
  *idx = startidx + blksize + 2 - 1;
  
  gcry_cipher_close(cipher_hd);
  
  return 0;
}

/**
 * \brief Parse literal data packet out of message
 *
 * The literal data is stored in an allocated buffer in |pkt|.  |idx| is
 * incremented to point to the last byte of literal data.
 *
 * It is the caller's responsibility to free the structure and its data
 * when finished.  This is handled in spgp_free_packet().
 *
 * \param msg Message to parse
 * \param idx Index to current location of |msg|, points to last byte of data
 *            in literal data packet when finished.
 * \param length Length of |msg|.
 * \param pkt PGP packet structure to fill with data.
 * \return 0 on success, non-zero on failure.
 */
static uint8_t spgp_parse_literal_packet(uint8_t *msg, 
                                         uint32_t *idx, 
                                          uint32_t length, 
                                         spgp_packet_t *pkt) {
  spgp_literal_pkt_t *literal = NULL;
  uint32_t date;
  uint32_t startidx;
  uint8_t format;
  
  LOG_PRINT("Parsing literal packet\n");

  if (NULL == msg || NULL == idx || NULL == pkt || length == 0)                              
    RAISE(INVALID_ARGS);

  startidx = *idx;

  pkt->c.literal = malloc(sizeof(*(pkt->c.literal)));
  if (NULL == pkt->c.literal) RAISE(OUT_OF_MEMORY);
  memset(pkt->c.literal, 0, sizeof(*(pkt->c.literal)));
  literal = pkt->c.literal;                                       

  // Read the format of the message.  This is ignored.
  format = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);

  // Read the length byte of the fylename
  literal->filenameLen = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  
  // Read the filename
  literal->filename = malloc(literal->filenameLen + 1);
  if (NULL == literal->filename) RAISE(OUT_OF_MEMORY);
  memcpy(literal->filename, msg+*idx, literal->filenameLen);
  literal->filename[literal->filenameLen] = '\0';
  *idx += literal->filenameLen - 1;
  SAFE_IDX_INCREMENT(*idx, length);
  
  // Read the timestamp.  This is ignored.
  memcpy(&date, msg+*idx, sizeof(date));
  *idx += 3;
  SAFE_IDX_INCREMENT(*idx, length);
  
  // Read the actual data in to buffer
  literal->dataLen = pkt->header->contentLength - (*idx - startidx);
  literal->data = malloc(literal->dataLen);
  if (NULL == literal->data) RAISE(OUT_OF_MEMORY);
  memcpy(literal->data, msg+*idx, literal->dataLen);
  *idx += literal->dataLen - 1;
  
  LOG_PRINT("Stored %u bytes\n", literal->dataLen);
  
  return 0;
}

