/*
 *  pubkey.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "simplepgp.h"
#include "packet_private.h"
#include "pubkey.h"
#include "mpi.h"

uint8_t spgp_parse_public_key(uint8_t *msg, uint32_t *idx, 
                              uint32_t length, spgp_packet_t *pkt) {
  spgp_public_pkt_t *pub;
  
  LOG_PRINT("Parsing public key.\n");

  // Make sure we have enough bytes remaining for parsing
  if (length - *idx < pkt->header->contentLength) RAISE(BUFFER_OVERFLOW);

  // Allocate public key if it doesn't already exist.  It might exist if
  // this packet is a secret key.
  if (!(pkt->c.pub)) {
    pkt->c.pub = malloc(sizeof(*(pkt->c.pub)));
    if (NULL == pkt->c.pub) RAISE(OUT_OF_MEMORY);
    memset(pkt->c.pub, 0, sizeof(*(pkt->c.pub)));
  }
  
  pub = pkt->c.pub;
  
  pub->version = msg[*idx]; 
  SAFE_IDX_INCREMENT(*idx, length);
  
  // First byte is the version.
  if (pub->version != 4) RAISE(FORMAT_UNSUPPORTED);
  
  // Next 4 bytes are big-endian 'key creation time'
  if (length - *idx < 4) RAISE(BUFFER_OVERFLOW);
  memcpy(&(pub->creationTime), msg+*idx, 4);
  *idx += 3; // this puts us on last byte of creation time
  SAFE_IDX_INCREMENT(*idx, length); // this goes to next byte (safely)

  // Next byte identifies asymmetric algorithm
  pub->asymAlgo = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  LOG_PRINT("Asymmetric algorithm: %d\n", pub->asymAlgo);
  
  // Read variable number of MPIs (depends on asymmetric algorithm), each
  // of which are variable size.
  spgp_read_all_public_mpis(msg, idx, length, pkt->c.pub);
  LOG_PRINT("Read %u MPIs\n", pub->mpiCount);
  
  // Create and store fingerprint for this packet
  spgp_generate_fingerprint(pkt);
  
  return 0;
}

uint8_t spgp_generate_fingerprint(spgp_packet_t *pkt) {
  uint16_t packetSize;
  uint8_t packetHeaderSize;
  spgp_mpi_t *curMpi;
  uint8_t targetMpiCount;
  gcry_md_hd_t md;
  unsigned char *hash;
  int i;
  
  if (NULL == pkt) RAISE(INVALID_ARGS);
  
  // Start with header info: 1 version, 4 creation time, 1 algorithm
  packetHeaderSize = sizeof(pkt->c.pub->version) +
    sizeof(pkt->c.pub->creationTime) +
    sizeof(pkt->c.pub->asymAlgo);
  packetSize = packetHeaderSize;

  // Figure out how many MPIs to add
  switch(pkt->c.pub->asymAlgo) {
    case ASYM_ALGO_RSA:
      targetMpiCount = 2;
      break;
    case ASYM_ALGO_DSA:
      targetMpiCount = 4;
      break;
    case ASYM_ALGO_ELGAMAL:
      targetMpiCount = 3;
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
  }
  
  // Add size of each MPI
  curMpi = pkt->c.pub->mpiHead;
  i = 0;
  while (curMpi && i < targetMpiCount) {
    packetSize += curMpi->count + 2; // add 2 for MPI header
    curMpi = curMpi->next;
    i++;
  }
      
  // Give data to hash to gcrypt
  if (gcry_md_open (&md, GCRY_MD_SHA1, 0) != 0) RAISE(GCRY_ERROR);
  gcry_md_putc(md, 0x99 );
  gcry_md_putc(md, packetSize >> 8);
  gcry_md_putc(md, packetSize);
  gcry_md_putc(md, pkt->c.pub->version);
  gcry_md_putc(md, pkt->c.pub->creationTime);
  gcry_md_putc(md, pkt->c.pub->creationTime >> 8);
  gcry_md_putc(md, pkt->c.pub->creationTime >> 16);
  gcry_md_putc(md, pkt->c.pub->creationTime >> 24);
  gcry_md_putc(md, pkt->c.pub->asymAlgo);
  
  // Write the public key MPIs
  curMpi = pkt->c.pub->mpiHead;
  i = 0;
  while (curMpi && i < targetMpiCount) {
    gcry_md_write(md, curMpi->data, curMpi->count + 2);
    curMpi = curMpi->next;
    i++;
  }
  
  // Perform SHA-1 hash
  gcry_md_final(md);
  hash = gcry_md_read(md, 0);

  // Copy hash results (20-bytes) into fingerprint
  pkt->c.pub->fingerprint = malloc(20);
  if (NULL == pkt->c.pub->fingerprint) RAISE(OUT_OF_MEMORY);
  memcpy(pkt->c.pub->fingerprint, hash, 20);
  
  LOG_PRINT("HASH: ");
  for (targetMpiCount=0; targetMpiCount < 20; targetMpiCount++) {
    fprintf(stderr, "%.2X", pkt->c.pub->fingerprint[targetMpiCount]);
  }
  fprintf(stderr,"\n");
  
  gcry_md_close(md);
  
  return 0;
}
