/*
 *  seckey.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "simplepgp.h"
#include "packet_private.h"
#include "seckey.h"
#include "mpi.h"
#include "util.h"
#include "pubkey.h"
#include "keychain.h"

uint8_t spgp_decrypt_all_secret_keys(spgp_packet_t *msg, 
                                     uint8_t *passphrase, uint32_t length) {
  spgp_packet_t *cur = msg;
  uint8_t err = 0;
  uint8_t haskey = 0;
  
  if (setjmp(exception)) {
      LOG_PRINT("Exception (0x%x)\n",_spgp_err);
      goto end;
  }

  if (NULL == msg || NULL == passphrase || length == 0) RAISE(INVALID_ARGS);

  while ((cur = spgp_next_secret_key_packet(cur)) != NULL) {
    LOG_PRINT("Decrypting secret key\n");
    spgp_decrypt_secret_key(cur, passphrase, length);
    cur = cur->next;
    haskey = 1;
  }
  
  // Add decrypted keys to keychain
  if (haskey)
    if (spgp_keychain_add_packet(msg) != 0) RAISE(KEYCHAIN_ERROR);
  
  end:
  return err;
}


uint8_t spgp_verify_decrypted_data(uint8_t *data, uint32_t length) {
  gcry_md_hd_t md;
  uint32_t hashlen = length - 20; // SHA1 hash is 20 bytes
  uint8_t *hashResult;
  int result;
  
  if (gcry_md_open (&md, GCRY_MD_SHA1, 0) != 0) RAISE(GCRY_ERROR);
  gcry_md_write(md, data, hashlen);
  gcry_md_final(md);
  hashResult = gcry_md_read(md, 0);
  if (NULL == hashResult) RAISE(GCRY_ERROR);
  result = memcmp(data+hashlen, hashResult, 20);
  gcry_md_close(md);
  return result;
}

/**
 * Generate cipher key to decrypt secret key packet
 *
 * The secret portion of a secret key packet can be encrypted with a 
 * symmetric cipher.  This function generates the 'key' that is used as
 * the input to the symmetric cipher.  This key is generated by hashing
 * a randomly generated salt, included in the packet, and the user's
 * passphrase (which must be provided).
 *
 * OpenPGP's standard allows multiple ways of generating the key by varying
 * the hash algorithm.
 *
 * @param pkt A secret key or secret subkey packet
 * @param passphrase User's passphrase to decrypt with
 * @param length Length (in bytes) of user's passphrase
 *
 * @return 0 on success.  Raises exception on error.
 *
 */
uint8_t spgp_generate_cipher_key(spgp_packet_t *pkt,
                                 uint8_t *passphrase, uint32_t length) {
  spgp_secret_pkt_t *secret;
  spgp_public_pkt_t *pub;
  gcry_md_hd_t md;
  uint32_t i;
  uint32_t keyBytesRemaining;// Bytes left to generate for key
  uint32_t hashLen;          // How long the hash is (algo-dependent)
  uint32_t hashIters;        // How many hash results to combine into key
  uint32_t hashBytes;        // Total number of bytes to hash each time
  uint32_t bufLen;           // Length of salt+passphrase
  uint32_t curHashCount;     // How many times we have performed full hash
  uint32_t hashCopies;       // How many integer copies of hashBuf per round
  uint32_t hashExtraBytes;   // How many extra bytes to hash for last round
  uint8_t *hashBuf;          // Store concatenated salt+passphrase
  uint8_t *hashResult;       // Store result of actual hash algorithm
  
  if (NULL == pkt || NULL == passphrase) RAISE(INVALID_ARGS);
  
  secret = pkt->c.secret;
  pub = pkt->c.pub;

  if (pkt->header->type != PKT_TYPE_SECRET_KEY &&
      pkt->header->type != PKT_TYPE_SECRET_SUBKEY)
      RAISE(INVALID_ARGS);
  
  // Determine how many bytes we need to produce for this cipher
  // Only supporting 3DES for this
  switch(secret->s2kEncryption) {
  case SYM_ALGO_3DES: secret->keyLength = 24; break;
  case SYM_ALGO_CAST5: secret->keyLength = 16; break;
  default: RAISE(FORMAT_UNSUPPORTED);
  }
   
  
  // Initialize hash algorithm, determine how many bytes produces per round
  switch (secret->s2kHashAlgo) {  
    case HASH_ALGO_SHA1:
      if (gcry_md_open (&md, GCRY_MD_SHA1, 0) != 0) RAISE(GCRY_ERROR);
      hashLen = 20;
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
      break;
  }
  
  // Determine how many times we have to hash to generate a large enough key
  // Ex: 3DES needs 24 bytes, SHA1 makes 20 bytes, so need to SHA1 hashes.
  hashIters = (secret->keyLength/hashLen) + (secret->keyLength%hashLen>0)?1:0;
  
  // What hashing mode to use.
  // Currently only supporting salted+iterated
  switch (secret->s2kSpecifier) {
    case S2K_TYPE_ITERATED:
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
      break;
  }
   
  // Allocate space for the key
  secret->key = malloc(secret->keyLength);
  if (NULL == secret->key) RAISE(OUT_OF_MEMORY);
  
  // Allocate a buffer to store the salt and passphrase combined
  // Since this buffer is local only, no exceptions can be raised after
  // this point or memory will be leaked.
  bufLen = secret->s2kSaltLength + length;
  hashBuf = malloc(bufLen);
  if (NULL == hashBuf) RAISE(OUT_OF_MEMORY);
  
  // Concatenate salt and passphrase into hashBuf
  memcpy(hashBuf, secret->s2kSalt, secret->s2kSaltLength);
  memcpy(hashBuf + secret->s2kSaltLength, passphrase, length);
  
  // Magic formula from RFC 4880.  This is number of bytes to hash over.
  hashBytes = (16 + (secret->s2kCount & 15)) << ((secret->s2kCount >> 4) + 6);
  
  // Figure out how many times to iterate over hashBuf to get hashBytes,
  // and how many extra bytes are needed at the end if not an even multiple.
  hashCopies = hashBytes / (bufLen);
  hashExtraBytes = hashBytes % (bufLen);
  
  keyBytesRemaining = secret->keyLength;
  
  // Loop until we have enough hash bytes to make the key
  curHashCount = 0;
  while (curHashCount <= hashIters && keyBytesRemaining) {
    for (i = 0; i < curHashCount; i++) {
      // pad front with 1 NUL byte per round (none on first round)
      gcry_md_putc(md, '\0'); 
    }
    // Copy the salt+passphrase combo into hash buffer as many times as fits
    for (i = 0; i < hashCopies; i++) {
      gcry_md_write(md, hashBuf, bufLen);
    }
    // Copy any leftover bytes into hash buffer to reach |hashBytes|
    if (hashExtraBytes) {
      gcry_md_write(md, hashBuf, hashExtraBytes);
    }
    // Perform the hash and append to the key
    gcry_md_final(md);
    hashResult = gcry_md_read(md, 0);
    
    if (keyBytesRemaining < hashLen) {
      memcpy(secret->key+(curHashCount*hashLen), 
             hashResult, 
             keyBytesRemaining);   
      keyBytesRemaining = 0;
    }
    else {
      memcpy(secret->key+(curHashCount*hashLen), 
             hashResult, 
             hashLen);
      keyBytesRemaining -= hashLen;
    }
    // Reset hash algorithm for next round
    gcry_md_reset(md);
    curHashCount++;
  }

  gcry_md_close(md);
  free(hashBuf);
  return 0;
}


uint8_t spgp_parse_secret_key(uint8_t *msg, uint32_t *idx, 
                              uint32_t length, spgp_packet_t *pkt) {
  spgp_secret_pkt_t *secret;
  spgp_public_pkt_t *pub;
  uint32_t startIdx = *idx;
  
  LOG_PRINT("Parsing secret key.\n");

  // Make sure we have enough bytes remaining for parsing
  if (length - *idx < pkt->header->contentLength) RAISE(BUFFER_OVERFLOW);

  // Allocate secret key in packet  
  pkt->c.secret = malloc(sizeof(*(pkt->c.secret)));
  if (NULL == pkt->c.secret) RAISE(OUT_OF_MEMORY);
  memset(pkt->c.secret, 0, sizeof(*(pkt->c.secret)));

  secret = pkt->c.secret;
  pub = pkt->c.pub;

  // Parse the public key section that starts it
  spgp_parse_public_key(msg, idx, length, pkt);
  // idx ends on last byte of public key.  One more to start secret key.
  SAFE_IDX_INCREMENT(*idx, length);
  
  // S2K Type byte tells how to (or if to) decrypt secret exponent
  secret->s2kType = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  switch (secret->s2kType) {
    case 0:
      // There is no encryption
      secret->s2kEncryption = 0;
      break;
    case 254:
    case 255:
      // Next byte is encryption type
      secret->s2kEncryption = msg[*idx];
      SAFE_IDX_INCREMENT(*idx, length);
      break;
    default:
      // This byte is encryption type
      secret->s2kEncryption = secret->s2kType;
      break;
  }
  LOG_PRINT("Encryption: %u\n", secret->s2kEncryption);
  
  if (secret->s2kEncryption) {
    // Secret exponent is encrypted (as it should be).  Time to decrypt.
    
    // S2K specifier tells us if there is a salt, and how to use it
    if (secret->s2kType >= 254) {
      secret->s2kSpecifier = msg[*idx];
      SAFE_IDX_INCREMENT(*idx, length);
      LOG_PRINT("S2K Specifier: %u\n", secret->s2kSpecifier);
    }
    
    // S2K hash algorithm specifies how to hash passphrase into a key
    secret->s2kHashAlgo = msg[*idx];
    SAFE_IDX_INCREMENT(*idx, length);
    LOG_PRINT("Hash algorithm: %u\n", secret->s2kHashAlgo);    
    
    // Read the salt if there is one
    switch (secret->s2kSpecifier) {
      case 1:
        spgp_read_salt(msg, idx, length, secret);
        break;
      case 3:
        spgp_read_salt(msg, idx, length, secret);
        // S2K Count is number of bytes to hash to make the key
        secret->s2kCount = msg[*idx];
        SAFE_IDX_INCREMENT(*idx, length);
        break;
      default:
        break;
    }
  }
  LOG_PRINT("Salt length: %u\n", secret->s2kSaltLength);
  
  // If it's not encrypted, we can just read the secret MPIs
  if (!secret->s2kEncryption) {
    spgp_read_all_secret_mpis(msg, idx, length, secret);
  }
  // If it is encrypted, just store it for now.  We'll decrypt later.
  else {
  
    // There's an initial vector (IV) here:
    spgp_read_iv(msg, idx, length, secret);
    LOG_PRINT("IV length: %u\n", secret->ivLength);
  
    // Figure out how much is left, and make sure it's available
    uint32_t packetOffset = *idx - startIdx;
    uint32_t remaining = pkt->header->contentLength - packetOffset;
    if (packetOffset >= pkt->header->contentLength) RAISE(BUFFER_OVERFLOW);
    
    // Allocate buffer and copy data
    secret->encryptedData = malloc(remaining);
    if (NULL == secret->encryptedData) RAISE(OUT_OF_MEMORY);
    memcpy(secret->encryptedData, msg+*idx, remaining);
    secret->encryptedDataLength = remaining;
    
    *idx += remaining-1;
    LOG_PRINT("Stored %u encrypted bytes.\n", remaining);
    // This is the end of the data, so we do NOT do a final idx increment
  }
  
  return 0;
}

spgp_packet_t *spgp_next_secret_key_packet(spgp_packet_t *msg) {
  spgp_packet_t *cur = msg;
  while (cur) {
    if (cur->header->type == PKT_TYPE_SECRET_KEY ||
        cur->header->type == PKT_TYPE_SECRET_SUBKEY)
        return cur;
    cur = cur->next;
  }
  return NULL;
}

uint8_t spgp_decrypt_secret_key(spgp_packet_t *pkt, 
                                uint8_t *passphrase, uint32_t length) {
  gcry_cipher_hd_t hd;
  spgp_secret_pkt_t *secret;
  spgp_public_pkt_t *pub;
  spgp_mpi_t *curMpi;
  uint32_t idx;
  uint32_t secretMpiCount;
  uint8_t *secdata;
  uint8_t i;
  uint8_t err = 0;

  if (NULL == pkt || NULL == passphrase || length == 0) RAISE(INVALID_ARGS);

  secret = pkt->c.secret;
  pub = pkt->c.pub;
  
  if (pkt->header->type != PKT_TYPE_SECRET_KEY &&
      pkt->header->type != PKT_TYPE_SECRET_SUBKEY)
      RAISE(INVALID_ARGS);

  if (secret->isDecrypted) return err; // already decrypted!
      
  spgp_generate_cipher_key(pkt, passphrase, length);

  switch (secret->s2kEncryption) {
    case SYM_ALGO_3DES:
    case SYM_ALGO_CAST5:
      if (gcry_cipher_open(&hd,
                           spgp_pgp_to_gcrypt_symmetric_algo(secret->s2kEncryption), 
                           GCRY_CIPHER_MODE_CFB, 
                           GCRY_CIPHER_SECURE | GCRY_CIPHER_ENABLE_SYNC) != 0)
        RAISE(GCRY_ERROR);
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
  }

  if (NULL == secret->key || NULL == secret->iv) RAISE(INCOMPLETE_PACKET);

  if (gcry_cipher_setkey(hd, secret->key, secret->keyLength) != 0)
    RAISE(GCRY_ERROR);
  if (gcry_cipher_setiv(hd, secret->iv, secret->ivLength) != 0)
    RAISE(GCRY_ERROR);
    
  // Allocate secret data memory.  Must free it before raising any exceptions!
  secdata = malloc(secret->encryptedDataLength);
  if (NULL == secdata) RAISE(OUT_OF_MEMORY);
  if (gcry_cipher_decrypt(hd, 
                          secdata, 
                          secret->encryptedDataLength, 
                          secret->encryptedData, 
                          secret->encryptedDataLength) != 0) {
    free(secdata);
    RAISE(GCRY_ERROR);
  }
  
  // Verify checksum
  if (spgp_verify_decrypted_data(secdata, secret->encryptedDataLength) != 0)
    RAISE(DECRYPT_FAILED);
  
  // Decode and store the secret MPIs (algo-specific):
  switch(pub->asymAlgo) {
    case ASYM_ALGO_RSA:
      secretMpiCount = 4;
      break;
    case ASYM_ALGO_DSA:
    case ASYM_ALGO_ELGAMAL:
      secretMpiCount = 1;
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
  }
    
  // Get to the last valid MPI
  if (NULL == pub->mpiHead) RAISE(INCOMPLETE_PACKET);
  curMpi = pub->mpiHead;
  while (curMpi->next) curMpi = curMpi->next;
  
  idx = 0;
  for (i = 0; i < secretMpiCount; i++) {
    curMpi->next = spgp_read_mpi(secdata, &idx, secret->encryptedDataLength);
    if (NULL == curMpi->next) RAISE(GENERIC_ERROR);
    curMpi = curMpi->next;
    SAFE_IDX_INCREMENT(idx, secret->encryptedDataLength);
    pub->mpiCount++;
  }
  secret->isDecrypted = 1;
  
  gcry_cipher_close(hd);
  free(secdata);
  
  end:
  return err;
}

uint8_t spgp_read_salt(uint8_t *msg, 
                       uint32_t *idx,
                       uint32_t length, 
                       spgp_secret_pkt_t *secret) {
  uint8_t saltLen = 0;
  
   if (NULL == msg || NULL == idx || 0 == length || NULL == secret)
    RAISE(INVALID_ARGS);
  
  if ((saltLen = spgp_salt_length_for_hash_algo(secret->s2kHashAlgo)) == 0) 
    RAISE(FORMAT_UNSUPPORTED);
  
  if (length - *idx < saltLen) RAISE(BUFFER_OVERFLOW);
  
  secret->s2kSalt = malloc(sizeof(*(secret->s2kSalt)) * saltLen);
  if (NULL == secret->s2kSalt) RAISE(OUT_OF_MEMORY);
  
  secret->s2kSaltLength = saltLen;
  memcpy(secret->s2kSalt, msg+*idx, saltLen);
  *idx += saltLen-1;
  SAFE_IDX_INCREMENT(*idx, length);
  
  return 0;
}

uint8_t spgp_read_iv(uint8_t *msg, 
                     uint32_t *idx,
                     uint32_t length, 
                     spgp_secret_pkt_t *secret) {
  uint8_t ivLen = 0;
  
   if (NULL == msg || NULL == idx || 0 == length || NULL == secret)
    RAISE(INVALID_ARGS);
  
  if ((ivLen = spgp_iv_length_for_symmetric_algo(secret->s2kEncryption)) == 0) 
    RAISE(FORMAT_UNSUPPORTED);
  
  if (length - *idx < ivLen) RAISE(BUFFER_OVERFLOW);
  
  secret->iv = malloc(sizeof(*(secret->iv)) * ivLen);
  if (NULL == secret->iv) RAISE(OUT_OF_MEMORY);
  
  secret->ivLength = ivLen;
  memcpy(secret->iv, msg+*idx, ivLen);
  *idx += ivLen-1;
  SAFE_IDX_INCREMENT(*idx, length);
  
  return 0;
}

/**
 * Find a specific secret key in a chain of packets.
 *
 * Finds a secret key in the given chain of packets, |chain|, with a key ID
 * matching |keyid|.  KeyID is the last 8 octets of the 64-octet key
 * fingerprint.
 *
 * @param chain Chain of packets containing at least one secret key
 * @param keyid 8-octet key ID
 * @return Packet containing matching secret key, or NULL if not found
 *
 */
spgp_packet_t *spgp_secret_key_matching_id(spgp_packet_t *chain,
                                           uint8_t *keyid) {
  spgp_packet_t *cur = NULL;
  
  if (NULL == chain || NULL == keyid) RAISE(INVALID_ARGS);
  
  cur = chain;
  while ((cur = spgp_next_secret_key_packet(cur)) != NULL) {
    if (memcmp((void*)((cur->c.pub->fingerprint)+12),keyid,8) == 0)
      return cur;
    cur = cur->next;
  }
  
  return NULL;
}



