/*
 *  session.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "session.h"
#include "simplepgp.h"
#include "packet_private.h"
#include "mpi.h"
#include "keychain.h"
#include "seckey.h"

spgp_packet_t *spgp_find_session_packet(spgp_packet_t *chain) {
  spgp_packet_t *cur;
  
  if (NULL == chain) RAISE(INVALID_ARGS);
  
  cur = chain;
  
  while (cur) {
    if (cur->header->type == PKT_TYPE_SESSION &&
        cur->c.session->key != NULL)
        return cur;
    cur = cur->prev;
  }
    
  return NULL;
}

uint8_t spgp_parse_session_packet(uint8_t *msg, uint32_t *idx, 
                                  uint32_t length, spgp_packet_t *pkt) {
  spgp_session_pkt_t *session;
  spgp_packet_t *key, *chain;
  gcry_sexp_t sexp_key, sexp_data, sexp_result;
  gcry_mpi_t mpis[10], mpi_result;
  spgp_mpi_t *cur;
  uint32_t checksum, sum;
  int i,mpi_count;
  unsigned long frame_len;
  uint8_t *frame;
  
  //TODO: Refactor this function, it's loong.
  
  LOG_PRINT("Parsing session packet.\n");

  if (NULL == msg || NULL == idx || length == 0 || NULL == pkt)
    RAISE(INVALID_ARGS);

  // Make sure we have enough bytes remaining for parsing
  if (length - *idx < pkt->header->contentLength) RAISE(BUFFER_OVERFLOW);

  // Allocate a session packet
  pkt->c.session = malloc(sizeof(*(pkt->c.session)));
  if (NULL == pkt->c.session) RAISE(OUT_OF_MEMORY);
  memset(pkt->c.session, 0, sizeof(*(pkt->c.session)));

  session = pkt->c.session;  
  
  session->version = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  LOG_PRINT("Version: %u\n", session->version);

  memcpy(session->keyid, msg+*idx, 8);
  *idx += 7;
  SAFE_IDX_INCREMENT(*idx, length);
  LOG_PRINT("Session for key ID: ");
  for (i = 0; i < 8; i++) fprintf(stderr, "%.2X",session->keyid[i]);
  fprintf(stderr, "\n");

  session->algo = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  
  // Read first MPI.  RSA only has one
  session->mpi1 = spgp_read_mpi(msg, idx, length);
  // Elgamal has a second MPI
  if (session->algo == ASYM_ALGO_ELGAMAL) {
    SAFE_IDX_INCREMENT(*idx, length);    
    session->mpi2 = spgp_read_mpi(msg, idx, length);
  }
  
  // DONE READING FROM STREAM AT THIS POINT
  // BELOW HERE -- DECRYPT SESSION KEY
  
  if (!spgp_keychain_is_valid()) RAISE(KEYCHAIN_ERROR);
  spgp_keychain_iter_start();
  while ((chain = spgp_keychain_iter_next()) != NULL) {
    if ((key = spgp_secret_key_matching_id(chain, session->keyid)) != NULL) {
      LOG_PRINT("Found a matching key in keychain.\n");
      break;
    }
  }
  spgp_keychain_iter_end();
  
  if (!key) return -1;
  
  
  for (cur=key->c.pub->mpiHead,i = 0; cur != NULL; cur = cur->next,i++) {
    gcry_mpi_scan (&(mpis[i]), GCRYMPI_FMT_PGP, cur->data, cur->count+2, NULL);
  }
  gcry_mpi_scan (&(mpis[i++]), GCRYMPI_FMT_PGP, 
                 session->mpi1->data, session->mpi1->count+2, NULL);
  if (session->mpi2) {
    gcry_mpi_scan (&(mpis[i++]), GCRYMPI_FMT_PGP, 
                   session->mpi2->data, session->mpi2->count+2, NULL);
  }
  mpi_count = i;

  
  switch (session->algo) {
    case ASYM_ALGO_RSA:
      gcry_sexp_build(&sexp_key, NULL,
        "(private-key(rsa(n%m)(e%m)(d%m)(p%m)(q%m)(u%m)))",
        mpis[0], mpis[1], mpis[2], mpis[3], mpis[4], mpis[5]);
      gcry_sexp_build (&sexp_data, NULL,
         "(enc-val(rsa(a%m)))", mpis[6]);
      gcry_pk_decrypt (&sexp_result, sexp_data, sexp_key);
      mpi_result = gcry_sexp_nth_mpi (sexp_result, 0, GCRYMPI_FMT_STD);
      break;
    case ASYM_ALGO_ELGAMAL:
      gcry_sexp_build(&sexp_key, NULL,
        "(private-key(elg(p%m)(g%m)(y%m)(x%m)))",
        mpis[0], mpis[1], mpis[2], mpis[3]);
      gcry_sexp_build (&sexp_data, NULL,
         "(enc-val(elg(a%m)(b%m)))", mpis[4], mpis[5]);
      gcry_pk_decrypt (&sexp_result, sexp_data, sexp_key);
      mpi_result = gcry_sexp_nth_mpi (sexp_result, 0, GCRYMPI_FMT_STD);
      break;
    default:
      RAISE(FORMAT_UNSUPPORTED);
  }

  if (!mpi_result) RAISE(GCRY_ERROR);

  gcry_mpi_print(GCRYMPI_FMT_PGP, NULL, 0, &frame_len, mpi_result);
  frame = malloc(frame_len);
  if (NULL == frame) RAISE(OUT_OF_MEMORY);
  gcry_mpi_print(GCRYMPI_FMT_PGP, frame, frame_len, NULL, mpi_result);

  if (mpi_result) {gcry_mpi_release(mpi_result);}
  if (sexp_key) {gcry_sexp_release(sexp_key);}
  if (sexp_data) {gcry_sexp_release(sexp_data);}
  if (sexp_result) {gcry_sexp_release(sexp_result);}

  for (i = 0; i < mpi_count; i++) {
    gcry_mpi_release(mpis[i]);
  }

  i = 2; // skip first two bytes, they're the length of the mpi
  if (frame[i++] != 2) RAISE(DECRYPT_FAILED);

  while (frame[i++] != 0 && i < frame_len) ; // Find the next 0 in frame
  
  // Algorithm is first byte after the 0
  session->symAlgo = frame[i];
  
  // Key length is determined from current index.  Drop 3 bytes: 1 for
  // algorithm, and 2 for the checksum at the end.
  session->keylen = frame_len - i - 3;
  i++;

  // Actual session key is the remaining bytes, except for the last two
  session->key = malloc(session->keylen);
  if (NULL == session->key) RAISE(OUT_OF_MEMORY);
  if (i+session->keylen >= frame_len) RAISE(DECRYPT_FAILED);
  memcpy(session->key, frame+i, session->keylen);

  // Checksum is last two bytes in buffer
  checksum = frame[frame_len-2]<<8 | frame[frame_len-1];
  
  // Verify checksum
  sum = 0;
  for (i = 0; i < session->keylen; i++) {
    sum = sum + (session->key[i] & 0xFF);
  }
  if (sum % 65536 != checksum) {
    LOG_PRINT("Session key checksum failed!\n");
    RAISE(DECRYPT_FAILED);
  }
  
  free(frame);
  frame = NULL;
  
  LOG_PRINT("Decrypted session key.\n");
  return 0;
}
