/*
 *  signature.c
 *  simplepgp
 *
 *  Created by Trevor Bentley on 11/26/11.
 *  Copyright 2011 Trevor Bentley. All rights reserved.
 *
 */

#include "packet_private.h"
#include "simplepgp.h"
#include "signature.h"
#include "mpi.h"
#include "util.h"
#include "keychain.h"

uint8_t spgp_parse_sig_subpackets(uint8_t *buf, 
                                  spgp_signature_pkt_t *sig,
                                  uint8_t isHashed) {
  uint16_t i; 
  uint16_t len;
  uint8_t pktlen, pkttype;
  
  if (NULL == buf || NULL == sig) RAISE(INVALID_ARGS);
  
  len = (isHashed)?sig->hashedSubLength:sig->unhashedSubLength;
  
  for (i = 0; i < len;) {
    pktlen = buf[i];
    pkttype = buf[i+1];
    switch (pkttype) {
    case SUBPKT_TYPE_ISSUER:
      memcpy(sig->issuer, buf+i+2, 8);
      LOG_PRINT("Signed with key: %X%X%X%X%X%X%X%X\n",
        buf[i+2],buf[i+3],buf[i+4],buf[i+5],
        buf[i+6],buf[i+7],buf[i+8],buf[i+9]);
    default:
      break;
    }
    i += pktlen;
  }
  
  return 0;
}


uint8_t spgp_parse_signature_packet(uint8_t *msg, 
                                    uint32_t *idx, 
                                    uint32_t length, 
                                    spgp_packet_t *pkt) {
  spgp_signature_pkt_t *sig = 0;
  spgp_literal_pkt_t *literal = 0;
  spgp_packet_t *pubkey = 0;
  gcry_md_hd_t md = 0;
  uint32_t startidx, stopidx;
  uint8_t *hash = 0;
  uint32_t totalLen;
  uint32_t mpibits;
  uint32_t i;
  gcry_mpi_t mpis[10] = {0};
  gcry_mpi_t hashmpi = 0;
  spgp_mpi_t *curMpi = 0;
  gcry_sexp_t sexp_key = 0;
  gcry_sexp_t sexp_hash = 0;
  gcry_sexp_t sexp_sig = 0;
  gcry_error_t gerr = 0;
  uint8_t mpiCount = 0;

  #define SPGP_PSP_ERROR() {gerr = 1; goto end;}

  LOG_PRINT("Parsing signature packet\n");
  
  if (msg == NULL || idx == NULL || pkt == NULL || 0 == length)
    RAISE(INVALID_ARGS);
    
  pkt->c.signature = malloc(sizeof(*(pkt->c.signature)));
  if (NULL == pkt->c.signature) RAISE(OUT_OF_MEMORY);
  memset(pkt->c.signature, 0, sizeof(*(pkt->c.signature)));
  sig = pkt->c.signature;

  startidx = *idx;

  sig->version = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);

  if (sig->version != 4) RAISE(FORMAT_UNSUPPORTED);

  sig->type = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);

  sig->asymAlgo = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);

  sig->hashAlgo = msg[*idx];
  SAFE_IDX_INCREMENT(*idx, length);
  
  LOG_PRINT("Signature type 0x%X, algo 0x%X, hash 0x%X\n",
    sig->type, sig->asymAlgo, sig->hashAlgo);
    
  sig->hashedSubLength = ((msg[*idx] & 0xFF) << 8) | msg[*idx + 1];
  *idx += 1;
  SAFE_IDX_INCREMENT(*idx, length);
  
  // skip hashed subpackets for now
  *idx += sig->hashedSubLength - 1;
  SAFE_IDX_INCREMENT(*idx, length);

  stopidx = *idx;

  sig->unhashedSubLength = ((msg[*idx] & 0xFF) << 8) | msg[*idx + 1];
  *idx += 1;
  SAFE_IDX_INCREMENT(*idx, length);
  
  spgp_parse_sig_subpackets(msg+*idx, sig, 0);
  
  // skip unhashed subpackets for now
  *idx += sig->unhashedSubLength - 1;
  SAFE_IDX_INCREMENT(*idx, length);

  sig->hashTest = ((msg[*idx] & 0xFF) << 8) | (msg[*idx + 1] & 0xFF);
  *idx += 1;
  SAFE_IDX_INCREMENT(*idx, length);

  sig->mpiHead = spgp_read_mpi(msg, idx, length);
  if (sig->asymAlgo == ASYM_ALGO_DSA) {
    SAFE_IDX_INCREMENT(*idx, length);
    sig->mpiHead->next = spgp_read_mpi(msg, idx, length);
  }

  // All data accounted for, idx incremented to the end
  // We can exit cleanly any time after this point

  if (NULL == pkt->prev || 
      pkt->prev->header->type != PKT_TYPE_LITERAL_DATA ||
      NULL == pkt->prev->c.literal ||
      NULL == pkt->prev->c.literal->data) 
      return -1;
      
  literal = pkt->prev->c.literal;

  if (gcry_md_open(&md, spgp_pgp_to_gcrypt_hash_algo(sig->hashAlgo), 0) != 0) 
    RAISE(GCRY_ERROR);
  gcry_md_write (md, literal->data, literal->dataLen);
  gcry_md_write (md, msg+startidx, stopidx-startidx);
  
  /* They hide this shit in here because they hate us.  What's really great
   is that the length will always be (hashedSubLength+6), and hashedSubLength
   is a 16-bit int, so in the worst case this final length would only be 3
   bytes.  Top byte in this 32-bit int will always be zero.
  
   RFC 4880 - 5.2.4
   V4 signatures also hash in a final trailer of six octets: the
   version of the Signature packet, i.e., 0x04; 0xFF; and a four-octet,
   big-endian number that is the length of the hashed data from the
   Signature packet (note that this number does not include these final
   six octets).*/
   
  totalLen = sig->hashedSubLength + 6;
   
  gcry_md_putc(md, sig->version);
  gcry_md_putc(md, 0xFF);
  gcry_md_putc(md, totalLen >> 24);
  gcry_md_putc(md, totalLen >> 16);
  gcry_md_putc(md, totalLen >>  8);
  gcry_md_putc(md, totalLen >>  0);
     
  gcry_md_final(md);
  hash = gcry_md_read(md, 0);
  
  if (memcmp(sig->issuer, "\0\0\0\0\0\0\0\0", 8) == 0)
    SPGP_PSP_ERROR();
    
  pubkey = spgp_keychain_public_key_with_id(sig->issuer);
  if (pubkey == NULL) SPGP_PSP_ERROR();
  
  for (curMpi = pubkey->c.pub->mpiHead, i=0; 
       curMpi != NULL; 
       curMpi = curMpi->next, i++) {
    gcry_mpi_scan (&(mpis[i]), GCRYMPI_FMT_PGP, curMpi->data, curMpi->count+2, NULL);
  }
  gcry_mpi_scan (&(mpis[i]), GCRYMPI_FMT_PGP, sig->mpiHead->data, sig->mpiHead->count+2, NULL);
  i++;
  if (sig->asymAlgo != ASYM_ALGO_RSA) {
    gcry_mpi_scan (&(mpis[i]), GCRYMPI_FMT_PGP, 
        sig->mpiHead->next->data, sig->mpiHead->next->count+2, NULL);
    i++;
  }
  mpiCount = i;
  
  switch (sig->asymAlgo) {
    case ASYM_ALGO_DSA:
      gcry_sexp_build(&sexp_key, NULL,
          "(public-key(dsa(p%m)(q%m)(g%m)(y%m)))",
          mpis[0], mpis[1], mpis[2], mpis[3]);
      gcry_sexp_build(&sexp_sig, NULL,
          "(sig-val(dsa(r%m)(s%m)))", 
          mpis[4], mpis[5]);
      break;
    case ASYM_ALGO_RSA:
      gcry_sexp_build (&sexp_key, NULL,
          "(public-key(rsa(n%m)(e%m)))",
          mpis[0], mpis[1]);
      gcry_sexp_build(&sexp_sig, NULL,
          "(sig-val(rsa(s%m)))", 
          mpis[2]);
      break;
    case ASYM_ALGO_ELGAMAL:
      gcry_sexp_build (&sexp_key, NULL,
          "(public-key(elg(p%m)(g%m)(y%m)))",
          mpis[0], mpis[1], mpis[2]);
      gcry_sexp_build(&sexp_sig, NULL,
          "(sig-val(elg(r%m)(s%m)))", 
          mpis[3], mpis[4]);
      break;
    default:
      SPGP_PSP_ERROR();
  }
  
  if (sig->asymAlgo == ASYM_ALGO_DSA) {
    mpibits = gcry_mpi_get_nbits(mpis[1]);
    if (mpibits % 8 != 0 || mpibits < 160) SPGP_PSP_ERROR();
    gcry_mpi_scan(&hashmpi, GCRYMPI_FMT_USG, hash, mpibits/8, NULL);
  }
  else {
    mpibits = gcry_mpi_get_nbits(mpis[0]);
    hashmpi = spgp_mpi_from_signature_hash(mpibits, sig->hashAlgo, hash);
  }
  
  
  
  gcry_sexp_build(&sexp_hash, NULL, "%m", hashmpi);
  gerr = gcry_pk_verify(sexp_sig, sexp_hash, sexp_key);
  if (!gerr) literal->validSig = 1;
  
  if (literal->validSig)
    LOG_PRINT("Signature validated successfully\n");
  else
    LOG_PRINT("WARNING: Invalid signature!\n");
  
  
end:  
  if (sexp_sig) gcry_sexp_release(sexp_sig);
  if (sexp_hash) gcry_sexp_release(sexp_hash);
  if (sexp_key) gcry_sexp_release(sexp_key);
  for (i = 0; i < mpiCount; i++) {
    gcry_mpi_release(mpis[i]);
  }
  if (hashmpi) gcry_mpi_release(hashmpi);
  if (md) gcry_md_close(md);
  if (gerr) RAISE(SIGNATURE_ERROR);
  
  return 0;
}
                                         
